# Quoins - A TurboGears blogging system.
# Copyright (C) 2008  James E. Blair <corvus@gnu.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from turbogears import controllers, expose, flash
from turbogears import identity, redirect
import turbogears as tg
from cherrypy import request, response
import cherrypy
import openid.consumer.consumer
import openid.server.server
import openid.extensions.sreg
from openid.store.sqlstore import MySQLStore
import MySQLdb
import sqlalchemy.engine.url
import types
import os.path

def flatten_params(d):
    r = {}
    for k,v in d.items():
        if isinstance(v, types.DictType):
            sub = flatten_params(v)
            for subk, subv in sub.items():
                if subk is None: subk = ''
                r[k+'.'+subk] = subv
        else:
            r[k] = v
    return r

def get_oid_connection():
    backupuri = tg.config.get('sqlalchemy.dburi')
    uri = tg.config.get('openid.store', backupuri)
    u = sqlalchemy.engine.url.make_url(uri)
    pw = u.password or ''
    conn = MySQLdb.connect (host = u.host,
                            user = u.username,
                            passwd = pw,
                            db = u.database)
    return conn

class OpenIDController(controllers.Controller):
    sreg_friendly = {
        'nickname': 'Nickname',
        'email': 'Email address',
        'fullname': 'Full name',
        'dob': 'Date of birth',
        'gender': 'Gender',
        'postcode': 'Postal code',
        'country': 'Country',
        'language': 'Language',
        'timezone': 'Time zone',
        }

    def url(self, obj=None):
        if obj is None:
            u = tg.url(self.path)
        elif isinstance(obj, basestring):
            if obj.startswith('/'): obj = obj[1:]
            u = tg.url(os.path.join(self.path, obj))
        return u

    def absolute_url(self, obj):
        return 'http://%s%s'%(tg.config.get('server.webhost'),
                              self.url(obj))
        
    def __init__(self, *args, **kw):
        self.get_name_from_id = kw.pop('get_name_from_id', lambda x: x)
        self.path = kw.pop('path', '/')
        super(OpenIDController, self).__init__(*args, **kw)

    def getSRegValue(self, user, field):
        val = None
        if field=='nickname': val = user.user_name
        if field=='fullname': val = user.display_name
        if field=='email': val = user.email_address
        return (field, self.sreg_friendly[field], val)

    def isAuthorized(self, id, trust_root):
        if identity.current.anonymous: return False
        name = self.get_name_from_id(id)
        if identity.current.user.user_name == name:
            # check to see if the user has set preferences for this trust
            # root, and return True.  also, sreg. XXX
            return False
        return False

    def _validate_identity(self, id):
        name = self.get_name_from_id(id)
        print name
        if identity.current.user.user_name != name:
            raise Exception("Not your ID")

    @identity.require(identity.not_anonymous())
    def showDecidePage(self, request):
        self._validate_identity(request.identity)

        cherrypy.session['oid_request']=request
        sreg_req = openid.extensions.sreg.SRegRequest.fromOpenIDRequest(request)
        user = identity.current.user

        sr_required = [self.getSRegValue(user,x) for x in sreg_req.required]
        sr_optional = [self.getSRegValue(user,x) for x in sreg_req.optional]

        return dict(tg_template="genshi:lemon.templates.oid_confirm",
                    request=request,
                    required = sr_required,
                    optional = sr_optional,
                    )

    @expose()
    def finish(self, **kw):
        request=cherrypy.session['oid_request']
        self._validate_identity(request.identity)        

        sreg_req = openid.extensions.sreg.SRegRequest.fromOpenIDRequest(request)
        user = identity.current.user

        store = MySQLStore(get_oid_connection())
        oserver = openid.server.server.Server(store,
                         self.absolute_url('/openid/server'))

        data = {}
        for field in sreg_req.required+sreg_req.optional:
            if kw.has_key(field):
                data[field]=self.getSRegValue(user, field)[2]

        sreg_resp = openid.extensions.sreg.SRegResponse.extractResponse(sreg_req, data)

        if kw.get('continue', None):
            response = request.answer(True)
            response.addExtension(sreg_resp)
        else:
            response = request.answer(False)
            
        webresponse = oserver.encodeResponse(response)

        cherrypy.response.status = webresponse.code
        cherrypy.response.headers.update(webresponse.headers)
        del cherrypy.session['oid_request']
        return webresponse.body

    @expose()
    def server(self, **kw):
        if (not identity.current.anonymous 
            and identity.was_login_attempted() 
            and not identity.get_identity_errors()
            and cherrypy.session.has_key('oid_request')):
            request=cherrypy.session['oid_request']
        else:
            kw = flatten_params(kw)
            request = None

        store = MySQLStore(get_oid_connection())
        oserver = openid.server.server.Server(store,
                         self.absolute_url('/openid/server'))
        if not request:
            request = oserver.decodeRequest(kw)

        if request.mode in ['checkid_immediate', 'checkid_setup']:
            if hasattr(request, 'returnToVerified'):
                if not (request.trustRootValid()
                        and request.returnToVerified()):
                    raise Exception("Not trusted")
            else:
                if not (request.trustRootValid()):
                    raise Exception("Not trusted")
            cherrypy.session['oid_request']=request
            if self.isAuthorized(request.identity, request.trust_root):
                response = request.answer(True)
            elif request.immediate:
                response = request.answer(False)
            else:
                cherrypy.request.params = {}
                return self.showDecidePage(request)
        else:
            response = oserver.handleRequest(request)

        webresponse = oserver.encodeResponse(response)

        cherrypy.response.status = webresponse.code
        cherrypy.response.headers.update(webresponse.headers)
        return webresponse.body

